from sandboxed.file_system import FileSystem
from sandboxed.internet_access import *
from sandboxed.specs import *
import time, os

class Sandbox:

    def __init__(self, logging=False):
        self.logging = logging
        return

    def _normalize(self, values):
        normalization = 0
        for value in values:
            # 0 value means, ABORT -> return 0
            if value == 0:
                return 0
            elif value == 1:
                normalization -= 1
            elif value == 2:
                normalization += 0.5
            elif value == 3:
                normalization += 2
            else:
                normalization += value

        normalization = normalization / len(values)
        return normalization


    def is_sandboxed(self):
        s1 = self._check_specs()
        s2 = self._check_internet()
        s3 = self._check_file_system()

        final_score = self._normalize([s1, s2, s3])

        percentage = int(100 - ((final_score / 5) * 100))

        if self.logging:
            self._printout(10, f"Conclusion: {percentage}% chance of being in a virtual environment.\n")

        return percentage / 100

    def _check_file_system(self):
        if self.logging:
            self._printout(10, "FILESYSTEM")

        s1, d1, e1 = FileSystem.check_vm_registry_keys()
        s2, d2, e2 = FileSystem.check_vm_files()
        s3, d3, e3 = FileSystem.check_vm_processes()

        # Indirectly trying to derive if it's an active valid user
        s4, d4, e4 = FileSystem.check_wifi_connections()
        s5, d5, e5 = FileSystem.check_application_files()
        s6, d6, e6 = FileSystem.check_prev_logins()

        if self.logging:
            self._printout(score=s1, description=d1, extra=e1)
            self._printout(s2, d2, e2)
            self._printout(s3, d3, e3)

            self._printout(s4, d4, e4)
            self._printout(s5, d5, e5)
            self._printout(s6, d6, e6)

        return self._normalize([s1, s2, s3, s4, s5, s6])


    def _check_internet(self):
        '''
        Since internet calls are being called synchronously could take a while,
        we will printout the status after every call made.
        Looks like shitty code, is shtty code, but gives user a nicer quicker printout.
        '''
        if self.logging:
            self._printout(10, "INTERNET ACCESS")

        s1, d1, e1 = InternetAccess.check_basic_ping()
        if self.logging:
            self._printout(score=s1, description=d1, extra=e1)

        s2, d2, e2 = InternetAccess.check_download_file()
        if self.logging:
            self._printout(s2, d2, e2)

        s3, d3, e3 = InternetAccess.check_http_post()
        if self.logging:
            self._printout(s3, d3, e3)

        s4, d4, e4 = InternetAccess.check_sockdnsreq()
        if self.logging:
            self._printout(s4, d4, e4)

        return self._normalize([s1, s2, s3, s4])

    def _check_specs(self):
        if self.logging:
            self._printout(10, "SPECS OF THE SYSTEM")

        s1, d1, e1 = Specs.check_hard_drive()
        s2, d2, e2 = Specs.check_ram_space()
        s3, d3, e3 = Specs.check_cpu_cores()

        s4, d4, e4 = Specs.check_serial_number()
        s5, d5, e5 = Specs.check_model()
        s6, d6, e6 = Specs.check_manufacturer()

        if self.logging:
            self._printout(score=s1, description=d1, extra=e1)
            self._printout(s2, d2, e2)
            self._printout(s3, d3, e3)

            self._printout(s4, d4, e4)
            self._printout(s5, d5, e5)
            self._printout(s6, d6, e6)

        return self._normalize([s1, s2, s3, s4, s5, s6])

    def _printout(self, score, description, extra=None):

        # enable color in cmd for Windows
        if os.name == 'nt':
            os.system('color')

        _CLEAR = '\033[0m'
        _BOLD = '\033[1m'

        _RED = '\033[91m'
        _ORANGE = '\033[33m'
        _LIME = '\033[92m'


        _ITALIC = '\033[3m'
        _UNDERLINED = '\033[4m'
        _YELLOW =  '\033[93m'

        # 10 is for headers of sections
        if score == 10:
            import shutil
            width = shutil.get_terminal_size((80, 20)).columns
            spaces = (width - len(description))
            if spaces % 2 == 0:
                left_spaces = int(spaces / 2)
                right_spaces = int(left_spaces)
            else:
                left_spaces = int((spaces-1) / 2)
                right_spaces = int(left_spaces + 1)

            left = f'{_BOLD}{_ITALIC}{_UNDERLINED}{_YELLOW}' + left_spaces * ' '
            right = right_spaces * ' ' + f'{_CLEAR}'
            description = f'{left}{description}{right}'

        if score == 5:
            description = f'{_BOLD}{_LIME}[+++++] {description}{_CLEAR}'
        elif score == 4:
            description = f'{_BOLD}{_LIME}[++++-] {description}{_CLEAR}'
        elif score == 3:
            description = f'{_BOLD}{_ORANGE}[+++--] {description}{_CLEAR}'
        elif score == 2:
            description = f'{_BOLD}{_ORANGE}[++---] {description}{_CLEAR}'
        elif score == 1:
            description = f'{_BOLD}{_RED}[+----] {description}{_CLEAR}'
        elif score == 0:
            description = f'{_BOLD}{_RED}[-----] {description}{_CLEAR}'

        # make it a bit more exciting
        print()
        print(description)
        time.sleep(0.2)
        if extra:
            print(f"-> {extra}")
            time.sleep(1)
        time.sleep(1.5)



import os
import psutil
import subprocess
import platform

# registry keys can only be checked for Windows
if os.name == 'nt':
# winreg is a built in library for windows python users, no need to pip install
    import winreg
    import win32net

'''
This class is mostly aimed for Windows
I don't know how linux based operating systems act in a virtual machine
'''
class FileSystem:

    _KEYS = [
        r"SOFTWARE\Oracle\VirtualBox Guest Additions",
        r"HARDWARE\ACPI\DSDT\VBOX__",
        r"HARDWARE\ACPI\FADT\VBOX__",
        r"HARDWARE\ACPI\RSDT\VBOX__",
        r"SYSTEM\ControlSet001\Services\VBoxGuest",
        r"SYSTEM\ControlSet001\Services\VBoxMouse",
        r"SYSTEM\ControlSet001\Services\VBoxService",
        r"SYSTEM\ControlSet001\Services\VBoxSF",
        r"SYSTEM\ControlSet001\Services\VBoxVideo",
        r"SOFTWARE\VMware, Inc.\VMware Tools",
    ]

    _FILES = [
        r"C:\WINDOWS\system32\drivers\VBoxMouse.sys",
        r"C:\WINDOWS\system32\drivers\VBoxGuest.sys",
        r"C:\WINDOWS\system32\drivers\VBoxSF.sys",
        r"C:\WINDOWS\system32\drivers\VBoxVideo.sys",
        r"C:\WINDOWS\system32\vboxdisp.dll",
        r"C:\WINDOWS\system32\vboxhook.dll",
        r"C:\WINDOWS\system32\vboxmrxnp.dll",
        r"C:\WINDOWS\system32\vboxogl.dll",
        r"C:\WINDOWS\system32\vboxoglarrayspu.dll",
        r"C:\WINDOWS\system32\vboxoglcrutil.dll",
        r"C:\WINDOWS\system32\vboxoglerrorspu.dll",
        r"C:\WINDOWS\system32\vboxoglfeedbackspu.dll",
        r"C:\WINDOWS\system32\vboxoglpackspu.dll",
        r"C:\WINDOWS\system32\vboxoglpassthroughspu.dll",
        r"C:\WINDOWS\system32\vboxservice.exe",
        r"C:\WINDOWS\system32\vboxtray.exe",
        r"C:\WINDOWS\system32\VBoxControl.exe",
        r"C:\WINDOWS\system32\drivers\vmmouse.sys",
        r"C:\WINDOWS\system32\drivers\vmhgfs.sys",
        r"C:\WINDOWS\system32\drivers\vmusbmouse.sys",
        r"C:\WINDOWS\system32\drivers\vmkdb.sys",
        r"C:\WINDOWS\system32\drivers\vmrawdsk.sys",
        r"C:\WINDOWS\system32\drivers\vmmemctl.sys",
        r"C:\WINDOWS\system32\drivers\vm3dmp.sys",
        r"C:\WINDOWS\system32\drivers\vmci.sys",
        r"C:\WINDOWS\system32\drivers\vmsci.sys",
        r"C:\WINDOWS\system32\drivers\vmx_svga.sys"
    ]

    _PROCESSES = [
        "vboxservices.exe",
        "vboxservice.exe",
        "vboxtray.exe",
        "xenservice.exe",
        "VMSrvc.exe",
        "vemusrvc.exe",
        "VMUSrvc.exe",
        "qemu-ga.exe",
        "prl_cc.exe",
        "prl_tools.exe",
        "vmtoolsd.exe",
        "df5serv.exe",
    ]

    @staticmethod
    def check_vm_registry_keys():
        if os.name != 'nt':
            return 5, "VM REGISTRY KEYS are None.", "This test can only be run on Windows. Considering this test successful."

        score = 5
        description = f"REGISTRY KEYS will look for VM related keys."
        explanation = None

        for key in FileSystem._KEYS:
            try:
                reg = winreg.ConnectRegistry(None, winreg.HKEY_LOCAL_MACHINE)
                winreg.OpenKey(reg, key)

                score = 0
                explanation = f"Found a key {key} which is related to VM."
            except:
                # no such key, which is good
                continue

        return score, description, explanation

    @staticmethod
    def check_vm_files():
        if os.name != 'nt':
            return 5, "VM RELATED FILES are None.", "This test can only be run on Windows. Considering this test successful."

        score = 5
        description = f"FILES will look for VM related files."
        explanation = None

        for filepath in FileSystem._FILES:
            if os.path.exists(filepath):
                score = 0
                explanation = f"Found a file {filepath} which is related to VM."

        return score, description, explanation

    @staticmethod
    def check_vm_processes():
        if os.name != 'nt':
            return 5, "VM RELATED PROCESSES are None.", "This test can only be run on Windows. Considering this test successful."

        score = 5
        description = f"PROCESSES will look for VM related processes."
        explanation = None

        try:
            for process in FileSystem._PROCESSES:
                for current_p in  psutil.process_iter():
                    if process.lower() == current_p.name().lower():
                        score = 0
                        explanation = f"Found a process: {process} which is related to VM."
                        break

        except psutil.AccessDenied:
            explanation = "Access to processes was denied. That's a good thing. Continuing."

        return score, description, explanation

    @staticmethod
    # only works for Windows
    def check_wifi_connections():
        '''
        This function will detect how many stored wifi access points the computer has.
        Ideally, a laptop has been connected on multiple wifi access points.
        If only been accessed to 1 wifi point that's fine (for example Desktop PC),
        but this must increase overall suspicion
        '''
        # Will increase to 1, if multiple wifi connections
        if os.name != 'nt':
            return 5, "STORED WIFI are None.", "This test can only be run on Windows. Considering this test successful."

        description = f"WIFI CONNECTIONS will look for stored wifi points."

        try:
            data = subprocess.check_output(['netsh', 'wlan', 'show', 'profiles']).decode('utf-8').split('\n')
            profiles = [i.split(":")[1][1:-1] for i in data if "All User Profile" in i]

            amount = len(profiles)

            explanation = f"Found {amount} stored wifi access points."
            if amount == 0:
                score = 1
            elif amount == 1:
                score = 3
            elif amount <= 5:
                score = 4
            else:
                score = 5

        except Exception as e:
            score = 3
            explanation = f"Something went wrong, this should not get benefit of the doubt.\nexception: {e}"

        return score, description, explanation

    @staticmethod
    def check_application_files():
        '''
        A 'normal' computer mostly has a lot of application files and data.
        If that's not the case, we're most likely inside a virtual machine.
        '''

        path = ''
        if os.name == 'nt':  # Windows
            path = os.getenv('localappdata')
        elif os.name == 'posix':
            path = os.getenv('HOME')
            if platform.system().lower() == 'darwin':  # MacOS
                path += '/Library/Application Support/'
            else:  # Linux: TODO: Don't use Linux so no idea where to look
                path += '/.config/'

        input_path = path
        amount = len(os.listdir(input_path))

        description = f"APPLICATION FILES will look for amount of stored files."
        explanation = f"Found {amount} files in {path}."

        if amount < 25:
            score = 1
            explanation = explanation + " Looks like a VM or a new PC."
        elif amount < 30:
            score = 2
            explanation = explanation + " Looks like a VM or a new PC."
        elif amount < 45:
            score = 3
            explanation = explanation + " Looks like a VM or rather unused PC."
        elif amount < 70:
            score = 4
        else:
            score = 5

        return score, description, explanation

    @staticmethod
    # only works for Windows
    def check_prev_logins():
        if os.name != 'nt':
            return 5, "PREV LOGINS are None.", "This test can only be run on Windows. Considering this test successful."


        users,nusers,_ = win32net.NetUserEnum(None,2)
        logons = 0
        for user in users:
            logons += int(user['num_logons'])

        description = f"PREV LOGINS will look for the amount of logins on the pc."
        explanation = f"Amount of logins on the pc is {logons}."

        if logons < 25:
            score = 1
            explanation = explanation + " Looks like a VM or a new PC."
        elif logons < 100:
            score = 2
            explanation = explanation + " Looks like a VM or a new PC."
        elif logons < 200:
            score = 3
            explanation = explanation + " Looks like a VM or rather unused PC."
        elif logons < 1000:
            score = 4
        else:
            score = 5

        return score, description, explanation



import psutil
import os
import subprocess
import math

class Specs:
    @staticmethod
    def check_hard_drive():
        pdisk = psutil.disk_usage('../')

        # put amount in gigabytes
        total = math.ceil(pdisk.total / (2 ** 30))
        used = pdisk.used / (2 ** 30)
        free = pdisk.free / (2 ** 30)
        percentage = pdisk.percent

        description = f"HARD DRIVE has a total storage of {total} GigaBytes."
        explanation = None

        # RED MASSIVE BAD
        if total < 60:
            score = 0
            explanation = "EVERY computer has at least 100 GiB."
        elif total < 100:
            score = 1
            explanation = "EVERY computer has at least 100 GiB."

        # ORANGE
        elif total < 200:
            score = 2
            explanation = "MOSTs computer have at least 200 GiB."
        elif total < 300:
            score = 3
            explanation = "AVERAGE computers have at least 300 GiB."

        # GREEN
        elif total < 500:
            score = 4 # almost perfect
        else:
            score = 5 # perfect score

        return score, description, explanation

    @staticmethod
    def check_ram_space():
        # get RAM and convert into gigabytes
        total = math.ceil(psutil.virtual_memory().total / (2 ** 30))

        description = f"RAM has a total storage of {total} GigaBytes."
        explanation = None

        # RED MASSIVE BAD
        if total < 2:
            score = 0
            explanation = "EVERY computer has at least 2 GiB."
        if total < 3:
            score = 1
            explanation = "MOST computers have at least 3 GiB."

        # ORANGE
        elif total < 4:
            score = 2
            explanation = "AVERAGE computer has at least 4 GiB."
        elif total < 8:
            score = 3
            explanation = "GOOD computers have at least 8 GiB."

        # GREEN
        elif total < 16:
            score = 4 # almost perfect
        else:
            score = 5 # perfect score

        return score, description, explanation

    @staticmethod
    def check_cpu_cores():
        # logical cores is most of the times double the physical
        logical_cores = psutil.cpu_count(logical=True) # most of the time double as physical
        physical_cores = psutil.cpu_count(logical=False)

        description = f"CPU has a total of {logical_cores} logical cores."
        explanation = None
        # if amount of cpu cores is uneven, then it's definitely a VM!
        # Most VM's have 1 CPU core, but all uneven amounts are never created by pc makers.

        # RED
        if logical_cores % 2 != 0:
            score = 0
            explanation = "EVERY computer should have an even amount of cpu cores."

        # ORANGE
        elif logical_cores < 4:
            score = 3
            explanation = "MOST computers have at least 4 logical cpu cores."

        # GREEN
        elif logical_cores < 8:
            score = 4
        else:
            score = 5

        return score, description, explanation

    @staticmethod
    def check_serial_number():
        # command only works if it's windows
        if os.name != 'nt':
            return 5, "SERIAL NUMBER is None.", "This test can only be run on Windows. Considering this test successful."

        serial = None
        explanation = None

        command = 'wmic bios get serialnumber'
        try:
            output = subprocess.check_output(command, shell=True)
            serial = output.decode().split('\n')[1].split(' ')[0]

            if str(serial) == "0":
                score = 0
                explanation = "Serial number CANNOT be 0 for a real computer."
            else:
                score = 5

        except Exception as e:
            score = 5
            explanation = f"Something went wrong, so giving benefit of the doubt. Considering this test successful.\nexception: {e}"

        description = f"SERIAL NUMBER is {serial}."

        return score, description, explanation

    # TODO add all known models of virtual machines
    _MODELS = [
        'virtualbox',
        'vmware'
    ]

    @staticmethod
    def check_model():
        # command only works if it's windows
        if os.name != 'nt':
            return 5, "MODEL is None.", "This test can only be run on Windows. Considering this test successful."

        model = None
        explanation = None

        command = 'wmic computersystem get model'
        try:
            output = subprocess.check_output(command, shell=True)
            model = output.decode().split('\n')[1].split(' ')[0]
            if model.lower() in Specs._MODELS:
                score = 0
                explanation = "MODEL has been linked to a virtual machine."
            else:
                score = 5

        except Exception as e:
            score = 5
            explanation = f"Something went wrong, so giving benefit of the doubt. Considering this test successful.\nexception: {e}"

        description = f"MODEL is {model}."

        return score, description, explanation

    # TODO add all known manufacturers of virtual machines
    _MANUFACTURER = [
        'innotek'
    ]

    @staticmethod
    def check_manufacturer():
        # command only works if it's windows
        if os.name != 'nt':
            return 5, "MANUFACTURER is None.", "This test can only be run on Windows. Considering this test successful."

        manufacturer = None
        explanation = None

        command = 'wmic computersystem get manufacturer'
        try:
            output = subprocess.check_output(command, shell=True)
            manufacturer = output.decode().split('\n')[1].split(' ')[0]
            if manufacturer.lower() in Specs._MANUFACTURER:
                score = 0
                explanation = "MANUFACTURER has been linked to a virtual machine."
            else:
                score = 5

        except Exception as e:
            score = 5
            explanation = f"Something went wrong, so giving benefit of the doubt. Considering this test successful.\nexception: {e}"

        description = f"MANUFACTURER is {manufacturer}."

        return score, description, explanation



import os
import subprocess
import requests
import socket

'''
This class will check whether there is 'real' connection to the internet.
'''
class InternetAccess:

    @staticmethod
    def check_basic_ping(amount=4, host='google.com'):
        if os.name == 'nt':
            # windows based
            command = f"ping -w {amount} {host}"
        else:
            # linux based
            command = f"ping -c {amount} {host}"

        description = f"BASIC PING makes {amount} pings to {host}."
        score = 5
        explanation = None

        try:
            # -> 64 bytes from 172.217.168.206: icmp_seq=0 ttl=119 time=21.761 ms
            output = subprocess.check_output(command, shell=True)

            # somehow check if ping is normal?
            for o in output.decode().split('\n'):
                if 'time=' in o:
                    latency = o.split('time=')[-1].split(" ms")[0]

        # if wifi is disabled, then output will be
        except subprocess.CalledProcessError as e:
            # -> 'ping: cannot resolve google.com: Unknown host'
            score = 3
            explanation = "BASIC PING indicates that your wifi is disabled."

        return score, description, explanation

    @staticmethod
    def check_download_file():

        description = f"DOWNLOAD FILE check if http get-request is possible."
        score = 5
        explanation = None

        try:
            r = requests.get('http://ipv4.download.thinkbroadband.com/20MB.zip')
            if r.status_code != 200:
                score = 3
                explanation = f"Status code is {r.status_code}, which is not 200 like expected."
        except Exception as e:
            score = 3
            explanation = f"Something went wrong, but should not give benefit of doubt.\nexception: {e}."

        return score, description, explanation

    @staticmethod
    def check_http_post():
        description = f"POST REQUEST check if http post-request is possible."
        score = 5
        explanation = None

        try:
            r = requests.post("http://bugs.python.org", data={'number': 12524, 'type': 'issue', 'action': 'show'})
            if r.status_code != 200:
                score = 3
                explanation = f"Status code is {r.status_code}, which is not 200 like expected."
        except Exception as e:
            score = 3
            explanation = f"Something went wrong, but should not give benefit of doubt.\nexception: {e}."

        return score, description, explanation

    @staticmethod
    def check_sockdnsreq():
        '''
        With sockets you go on the level lower and actually control the connection and send/receive raw bytes. 
        HTTP connection is a protocol that runs on a socket. 
        HTTP connection is a higher-level abstraction of a network connection.
        '''
        description = f"SOCKET REQUEST to check if lower level socket dns requests are possible."
        score = 5
        explanation = None

        try:
            addr1 = socket.gethostbyname('google.com') # 172.217.168.206
            socket.gethostbyname('yahoo.com')
            socket.gethostbyname('facebook.com')
            socket.gethostbyname('instagram.com')
        except Exception as e:
            score = 3
            explanation = f"Something went wrong, but should not give benefit of doubt.\nexception: {e}."

        return score, description, explanation



import setuptools
from distutils.core import setup

setup(
  name='sandboxed',         # How you named your folder
  packages=['sandboxed'],   # Chose the same as "name"
  version='2.0',      # Start with a small number and increase it with every change you make
  license='MIT',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  description='Sandbox/Virtual Machine detection',   # Give a short description about your library
  author='Frederik Mees',                   # Type in your name
  author_email='frederik.mees@gmail.com',      # Type in your E-Mail
  url='https://github.com/frederikme/sandbox-evasion',   # Provide either the link to your github or to your website
  download_url='https://github.com/frederikme/sandbox-evasion/archive/2.0.tar.gz',    # I explain this later on
  keywords=['Sandbox', 'Virtual machine', 'Virtualbox', 'detection'],   # Keywords that define your project best
  install_requires=['psutil', 'requests', 'pypiwin32; platform_system == "Windows"'],

  package_data={},

  classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your loadingz
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
  ],
)
